"""All the visual effects that can be applied to VideoClip."""

# import every video fx function

from moviepy.video.fx.AccelDecel import AccelDecel
from moviepy.video.fx.BlackAndWhite import BlackAndWhite
from moviepy.video.fx.Blink import Blink
from moviepy.video.fx.Crop import Crop
from moviepy.video.fx.CrossFadeIn import CrossFadeIn
from moviepy.video.fx.CrossFadeOut import CrossFadeOut
from moviepy.video.fx.EvenSize import EvenSize
from moviepy.video.fx.FadeIn import FadeIn
from moviepy.video.fx.FadeOut import FadeOut
from moviepy.video.fx.Freeze import Freeze
from moviepy.video.fx.FreezeRegion import FreezeRegion
from moviepy.video.fx.GammaCorrection import GammaCorrection
from moviepy.video.fx.HeadBlur import HeadBlur
from moviepy.video.fx.InvertColors import InvertColors
from moviepy.video.fx.Loop import Loop
from moviepy.video.fx.LumContrast import LumContrast
from moviepy.video.fx.MakeLoopable import MakeLoopable
from moviepy.video.fx.Margin import Margin
from moviepy.video.fx.MaskColor import MaskColor
from moviepy.video.fx.MasksAnd import MasksAnd
from moviepy.video.fx.MasksOr import MasksOr
from moviepy.video.fx.MirrorX import MirrorX
from moviepy.video.fx.MirrorY import MirrorY
from moviepy.video.fx.MultiplyColor import MultiplyColor
from moviepy.video.fx.MultiplySpeed import MultiplySpeed
from moviepy.video.fx.Painting import Painting
from moviepy.video.fx.Resize import Resize
from moviepy.video.fx.Rotate import Rotate
from moviepy.video.fx.Scroll import Scroll
from moviepy.video.fx.SlideIn import SlideIn
from moviepy.video.fx.SlideOut import SlideOut
from moviepy.video.fx.SuperSample import SuperSample
from moviepy.video.fx.TimeMirror import TimeMirror
from moviepy.video.fx.TimeSymmetrize import TimeSymmetrize


__all__ = (
    "AccelDecel",
    "BlackAndWhite",
    "Blink",
    "Crop",
    "CrossFadeIn",
    "CrossFadeOut",
    "EvenSize",
    "FadeIn",
    "FadeOut",
    "Freeze",
    "FreezeRegion",
    "GammaCorrection",
    "HeadBlur",
    "InvertColors",
    "Loop",
    "LumContrast",
    "MakeLoopable",
    "Margin",
    "MasksAnd",
    "MaskColor",
    "MasksOr",
    "MirrorX",
    "MirrorY",
    "MultiplyColor",
    "MultiplySpeed",
    "Painting",
    "Resize",
    "Rotate",
    "Scroll",
    "SlideIn",
    "SlideOut",
    "SuperSample",
    "TimeMirror",
    "TimeSymmetrize",
)
